import sys
import os
import json
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import datetime

import argparse
import logging

from time import sleep

from pymso5000.mso5000 import MSO5000

def parseArguments():
    ap = argparse.ArgumentParser(description = "MSO5000 data fetcher")
    ap.add_argument("-o", "--out", type=str, required=False, default=None, action='append', nargs="*", help="Add npz, svg or png output file (determined by extension")

    ap.add_argument("-m", "--host", type=str, required=False, default=None, help="Specify the hostname or IP address of the MSO. If not specified will be loaded from the configuration file")
    ap.add_argument("-p", "--port", type=int, required=False, default=5555, help="Network port of the MSO (default 5555)")

    ap.add_argument("-s", "--stat", type=str, required=False, default=None, action='append', nargs="*", help="Statistics to evaluate (mean)")
    ap.add_argument("-v", "--show", action="store_true", help="Show the plot using matplotlib")

    ap.add_argument("--endis", action="store_true", help="Enable requested channels, disable others")
    ap.add_argument("-d", "--differential", action="store_true", help="Perform a differential measurement by first aquiring the background, then the foreground and performing subtraction")
    ap.add_argument("--delay", type=int, required=False, default=10, help="Delay for background measurement (typical 10 seconds)")
    ap.add_argument("--noplot", type=int, required=False, default=None, action='append', nargs="*", help="Disable plotting for the specified channel")

    ap.add_argument("--plottitle", type=str, required=False, default="MSO5000", help="Set plot title")
    ap.add_argument("--xlabel", type=str, required=False, default=None, help="Timebase axis label")
    ap.add_argument("--ylabel", type=str, required=False, default=None, help="Y axis label")
    ap.add_argument("--ylabeld", type=str, required=False, default=None, help="Difference Y axis label")

    ap.add_argument("--loglevel", type=str, required=False, default="ERROR", help="Loglevel (CRITICAL, ERROR, WARNING, INFO, DEBUG)")
    ap.add_argument("--logfile", type=str, required=False, default=None, help="Log into the specified logfile instead of the console")

    ap.add_argument("--noautoscale", action = "store_true", help="Disable automatic scaling")

    ap.add_argument("rest", nargs=argparse.REMAINDER)

    args = ap.parse_args()

    # Validate we know the filetypes of -o if any
    if args.out is not None:
        for fname in args.out:
            if len(fname) < 1:
                print(f"Invalid output filename {fname} supplied")
                sys.exit(1)
            fname = fname[0]
            if len(fname) < 4:
                print(f"Invalid output filename {fname} supplied")
                sys.exit(1)
            if fname[-4:] not in [ ".png", ".npz", ".svg" ]:
                print(f"File format of {fname} not known by extension")
                sys.exit(1)

    loglvls = {
        "DEBUG" : logging.DEBUG,
        "INFO" : logging.INFO,
        "WARNING" : logging.WARNING,
        "ERROR" : logging.ERROR,
        "CRITICAL" : logging.CRITICAL
    }
    if not args.loglevel.upper()in loglvls:
        print("Unknown loglevel {args.loglevel}")
        sys.exit(1)

    logger = logging.getLogger()
    logger.setLevel(loglvls[args.loglevel.upper()])
    if args.logfile is not None:
        logger.addHandler(logging.FileHandler(args.logfile))
        logger.addHandler(logging.StreamHandler(sys.stderr))
    else:
        logger.addHandler(logging.StreamHandler(sys.stderr))

    return (args, logger)

def loadConfigFile(cfname = ".config/mso5000.cfg"):
    cfgPath = os.path.join(Path.home(), cfname)
    try:
        with open(cfgPath) as cfgFile:
            cfg = json.load(cfgFile)
    except FileNotFoundError:
        return None
    except JSONDecodeError as e:
        print(f"Failed to load configuration file {cfgname}")
        print(e)
        sys.exit(1)

    return cfg

def getScaleFactorAndPrefix(v):
    prefixlist = [
        (1e12, "T"),
        (1e9, "G"),
        (1e6, "M"),
        (1e3, "k"),
        (1, ""),
        (1e-3, "m"),
        (1e-6, "u"),
        (1e-9, "n"),
        (1e-12, "p"),
        (1e-15, "f")
    ]

    for i in range(len(prefixlist)):
        if v > prefixlist[i][0]:
            return (1.0 / prefixlist[i][0], prefixlist[i][1])
    return (1.0 / prefixlist[-1][0], prefixlist[-1][1])

def main():
    args, logger= parseArguments()
    cfgFile = loadConfigFile()

    if len(args.rest) < 1:
        print(f"No channels specified. See usage with {sys.argv[0]} --help")
        sys.exit(1)

    if (args.host is None) and (cfgFile is not None):
        if "host" in cfgFile:
            args.host = cfgFile["host"]
        if "port" in cfgFile:
            args.port = cfgFile["port"]

    # Now conenct to the MSO
    logger.debug(f"Connecting to {args.host} at port {args.port}")
    with MSO5000(
        address = args.host,
        port = args.port,
        useNumpy = True
    ) as mso:
        logger.debug("Connected")

        # Enable the requested channels, disable the others
        chRequested = []
        chNoplot = []
        requestedStats = []

        for ch in args.rest:
            try:
                nch = int(ch)
                if (nch < 1) or (nch > 4):
                    print(f"Invalid channel number {nch}")
                    sys.exit(1)
                chRequested.append(nch - 1)
            except ValueError:
                print(f"Invalid channel number {ch}")
                sys.exit(1)
        if args.noplot is not None:
            for ch in args.noplot:
                try:
                    nch = int(ch[0])
                    if (nch < 1) or (nch > 4):
                        print(f"Invalid channel number {nch}")
                        sys.exit(1)
                    chNoplot.append(nch - 1)
                except ValueError:
                    print(f"Invalid channel number {ch}")
                    sys.exit(1)
        if args.stat is not None:
            for st in args.stat:
                if st[0] in [ "mean", "fft", "ifft", "correlate", "autocorrelate" ]:
                    requestedStats.append(st[0])
                else:
                    print(f"Unknown statistics {st}")

        if args.endis:
            for i in range(4):
                if i in chRequested:
                    logger.info(f"Enabling channel {i+1}")
                    mso.set_channel_enable(i, True)
                else:
                    logger.info(f"Disabling channel {i+1}")
                    mso.set_channel_enable(i, False)

        #Query the waveform ...
        logger.info("Gathering data")
        data = mso.query_waveform(chRequested, stats = requestedStats)
        dataBG = None
        dataDiff = None

        if args.differential:
            logger.info(f"Sleeping for {args.delay} seconds till gathering background data")
            sleep(args.delay)

            logger.info("Gathering background data")
            dataBG = mso.query_waveform(chRequested, stats = requestedStats)

            dataDiff = {}
            for ch in chRequested:
                dataDiff[f"y{ch}"] = data[f"y{ch}"] - dataBG[f"y{ch}"]

        # Plot figures if a plot is requested or show has been enabled
        doPlot = False
        if args.show:
            doPlot = True
        if args.out is not None:
            for fname in args.out:
                fname = fname[0]
                if fname[-4:] in [ ".svg", ".png" ]:
                    doPlot = True
                if fname[-4:] in [ ".npz" ]:
                    # Store npz
                    np.savez(fname, data = data, background = dataBG, diffdata = dataDiff)
                    logger.info(f"Stored {fname}")
        else:
            # Always plot if we don't write output files ...
            doPlot = True

        if doPlot:
            # Determine the scaling factors for all data ...
            maxYval = 0
            for ich in chRequested:
                if ich not in chNoplot:
                    newmax = np.max(data[f"y{ich}"])
                    newmin = np.min(data[f"y{ich}"])
                    maxYval = np.max((newmax, maxYval, np.abs(newmin)))
                    if dataDiff is not None:
                        newmax = np.max(dataBG[f"y{ich}"])
                        newmin = np.min(dataBG[f"y{ich}"])
                        maxYval = np.max((newmax, maxYval, np.abs(newmin)))
            sfY, sfYLabel = getScaleFactorAndPrefix(maxYval)

            if dataDiff is not None:
                maxYDval = 0
                for ich in chRequested:
                    if ich not in chNoplot:
                        newmax = np.max(dataDiff[f"y{ich}"])
                        newmin = np.min(dataDiff[f"y{ich}"])
                        maxYDval = np.max((newmax, np.abs(newmin), maxYDval))
                sfYD, sfYDLabel = getScaleFactorAndPrefix(maxYDval)

            sfX, sfXLabel = getScaleFactorAndPrefix(np.max(data["x"]))

            if args.noautoscale:
                sfX, sfXLabel = 1, ""
                sfY, sfYLabel = 1, ""
                sfYD, sfYDLabel = 1, ""

            xlabel = "Time "
            ylabel = "Signal "
            ylabeld = "Difference signal "

            if args.xlabel:
                xlabel = f"{args.xlabel} "
            if args.ylabel:
                ylabel = f"{args.ylabel} "
            if args.ylabeld:
                ylabeld = f"{args.ylabeld} "

            nplt = 1
            if dataDiff is not None:
                nplt = 2

            fig, ax = plt.subplots(nplt, figsize=(6.4, 4.8 * nplt))

            fig.suptitle(args.plottitle)

            if nplt == 1:
                for ich in chRequested:
                    if ich not in chNoplot:
                        ax.plot(data["x"] * sfX, data[f"y{ich}"] * sfY, label = f"Channel {ich+1}")
                        if dataBG is not None:
                            ax.plot(dataBG["x"] * sfX, dataBG[f"y{ich}"] * sfY, label = f"Background channel {ich+1}")
                ax.set_xlabel(f"{xlabel}[{sfXLabel}s]")
                ax.set_ylabel(f"{ylabel}[{sfYLabel}V]")
                ax.grid()
                ax.legend()
            else:
                for ich in chRequested:
                    if ich not in chNoplot:
                        ax[0].plot(data["x"] * sfX, data[f"y{ich}"] * sfY, label = f"Channel {ich+1}")
                        if dataBG is not None:
                            ax[0].plot(data["x"] * sfX, data[f"y{ich}"] * sfY, label = f"Background channel {ich+1}")
                ax[0].set_xlabel(f"{xlabel}[{sfXLabel}s]")
                ax[0].set_ylabel(f"{ylabel}[{sfYLabel}V]")
                ax[0].grid()
                ax[0].legend()

                if dataDiff is not None:
                    for ich in chRequested:
                        if ich not in chNoplot:
                            ax[1].plot(data["x"] * sfX, dataDiff[f"y{ich}"] * sfYD, label = f"Difference channel {ich+1}")
                    ax[1].set_xlabel(f"{xlabel}[{sfXLabel}s]")
                    ax[1].set_ylabel(f"{ylabeld}[{sfYLabel}V]")
                    ax[1].grid()
                    ax[1].legend()

            if args.out is not None:
                for fn in args.out:
                    fname = fn[0]
                    if fname[-4:] in [ ".svg", ".png" ]:
                        plt.savefig(fname)
                        logger.info(f"Written plot {fname}")

            if args.show or (args.out is None):
                plt.show()
                

if __name__ == "__main__":
    main()



from labdevices.exceptions import CommunicationError_ProtocolViolation
from labdevices.exceptions import CommunicationError_Timeout
from labdevices.exceptions import CommunicationError_NotConnected

from labdevices.oscilloscope import Oscilloscope, OscilloscopeSweepMode, OscilloscopeTriggerMode, OscilloscopeTimebaseMode, OscilloscopeRunMode
from labdevices.scpi import SCPIDeviceEthernet
import atexit

from time import sleep

import socket

import logging
import datetime

class MSO5000(Oscilloscope):
    def __init__(
        self,

        address=None,
        port=5555,

        useNumpy = False
    ):
        self._scpi = SCPIDeviceEthernet(address, port, None)

        super().__init__(
            nChannels = 4,
            supportedSweepModes = [
                OscilloscopeSweepMode.AUTO,
                OscilloscopeSweepMode.NORMAL,
                OscilloscopeSweepMode.SINGLE
            ],
            supportedTriggerModes = [
                OscilloscopeTriggerMode.EDGE,
                OscilloscopeTriggerMode.PULSE,
                OscilloscopeTriggerMode.SLOPE
            ],
            supportedTimebaseModes = [
                OscilloscopeTimebaseMode.MAIN,
                OscilloscopeTimebaseMode.XY,
                OscilloscopeTimebaseMode.ROLL
            ],
            supportedRunModes = [
                OscilloscopeRunMode.STOP,
                OscilloscopeRunMode.RUN,
                OscilloscopeRunMode.SINGLE
            ],
            timebaseScale = (5e-9, 1000.0),
			voltageScale = (500e-6, 10),
            triggerForceSupported = True
        )

        self._probe_ratios = [ 1, 1, 1, 1 ]

        self._use_numpy = useNumpy

        atexit.register(self.__close)

    # Connection handling

    def _connect(self, address = None, port = None):
        if self._scpi.connect(address, port):
            # Ask for identity and verify ...
            idnString = self._idn()
            if not idnString.startswith("RIGOL TECHNOLOGIES,MSO50"):
                self._disconnect()
                raise ValueError(f"Unsupported device, identifies as {idnString}")

            idnParts = idnString.split(",")
            self._id = {
                'manufacturer' : idnParts[0],
                'product'      : idnParts[1],
                'serial'       : idnParts[2],
                'version'      : idnParts[3]
            }
        return True

    def _disconnect(self):
        self._scpi.disconnect()

    def _isConnected(self):
        return self._scpi.isConnected()

    # Context management

    def __enter__(self):
        if self._usedConnect:
            raise ValueError("Cannot use context management (with) on a connected port")

        # Run our internal connect method ...
        self._connect()

        self._usesContext = True
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__close()
        self._usesContext = False
    def __close(self):
        atexit.unregister(self.__close)
        if self._scpi.isConnected():
            self._off()
            self._disconnect()

    # Commands

    def _idn(self):
        return self._scpi.scpiQuery("*IDN?")

    def _identify(self):
        resp = self._idn()
        if resp is None:
            return None

        idnParts = resp.split(",")

        return {
            'manufacturer' : idnParts[0],
            'product'      : idnParts[1],
            'serial'       : idnParts[2],
            'version'      : idnParts[3]
        }

    def _off(self):
        pass

    def _set_channel_enable(self, channel, enabled):
        if (channel < 0) or (channel > 3):
            raise ValueError("Invalid channel number for MSO5000")
        if enabled:
            self._scpi.scpiCommand(f":CHAN{channel+1}:DISP ON")
        else:
            self._scpi.scpiCommand(f":CHAN{channel+1}:DISP OFF")

    def _is_channel_enabled(self, channel):
        if (channel < 0) or (channel > 3):
            raise ValueError("Invalid channel number for MSO5000")

        resp = self._scpi.scpiQuery(f":CHAN{channel+1}:DISP?")
        try:
            resp = int(resp)
            if resp == 1:
                return True
            elif resp == 0:
                return False
        except:
            pass

        raise CommunicationError_ProtocolViolation("Failed to query enabled status of channel")

    def _set_sweep_mode(self, mode):
        if mode == OscilloscopeSweepMode.AUTO:
            self._scpi.scpiCommand(":TRIG:SWE AUTO")
        elif mode == OscilloscopeSweepMode.NORMAL:
            self._scpi.scpiCommand(":TRIG:SWE NORM")
        elif mode == OscilloscopeSweepMode.SINGLE:
            self._scpi.scpiCommand(":TRIG:SWE SING")
        else:
            raise ValueError(f"Unknown sweep mode {mode} passed")

    def _get_sweep_mode(self):
        resp = self._scpi.scpiCommand(f":TRIG:SWE?")

        modes = {
            "NORM" : OscilloscopeSweepMode.NORMAL,
            "SING" : OscilloscopeSweepMode.SINGLE,
            "AUTO" : OscilloscopeSweepMode.AUTO
        }
        if resp in modes:
            return modes[resp]
        else:
            raise CommunicationError_ProtocolViolation(f"Unknown sweep mode {resp} received from device")


    def _set_trigger_mode(self, mode):
        if mode == OscilloscopeTriggerMode.EDGE:
            self._scpi.scpiCommand(":TRIG:MODE EDGE")
        elif mode == OscilloscopeTriggerMode.PULSE:
            self._scpi.scpiCommand(":TRIG:MODE PULS")
        elif mode == OscilloscopeTriggerMode.SLOPE:
            self._scpi.scpiCommand(":TRIG:MODE SLOP")

    def _get_trigger_mode(self):
        resp = self._scpi.scpiQuery(f":TRIG:MODE?")

        modes = {
            "EDGE" : OscilloscopeTriggerMode.EDGE,
            "PULS" : OscilloscopeTriggerMode.PULSE,
            "SLOP" : OscilloscopeTriggerMode.SLOPE
        }
        if resp in modes:
            return modes[resp]
        else:
            raise CommunicationError_ProtocolViolation(f"Unknown trigger mode {resp} received from device")

    def _force_trigger(self):
        self._scpi.scpiCommand(":TFOR")

    def _set_run_mode(self, mode):
        if mode == OscilloscopeRunMode.STOP:
            self._scpi.scpiCommand(":STOP")
        elif mode == OscilloscopeRunMode.SINGLE:
            self._scpi.scpiCommand(":SING")
        elif mode == OscilloscopeRunMode.RUN:
            self._scpi.scpiCommand(":RUN")

    def _get_run_mode(self, mode):
        resp = self._scpi.scpiQuery(":TRIG:STAT?")

        if resp == "STOP":
            return OscilloscopeRunMode.STOP
        elif (resp == "RUN") or (resp == "AUTO"):
            return OscilloscopeRunMode.RUN
        elif resp == "WAIT":
            return OscilloscopeRunMode.RUN

    def _set_timebase_mode(self, mode):
        modestr = {
            OscilloscopeTimebaseMode.MAIN : "MAIN",
            OscilloscopeTimebaseMode.XY   : "XY",
            OscilloscopeTimebaseMode.ROLL : "ROLL"
        }

        if mode not in modestr:
            raise ValueError(f"Unsupported timebase mode {mode}")

        self._scpi.scpiCommand(f":TIM:MODE {modestr[mode]}")

    def _get_timebase_mode(self):
        resp = self._scpi.scpiQuery(f":TIM:MODE?")

        modes = {
            "MAIN" : OscilloscopeTimebaseMode.MAIN,
            "XY"   : OscilloscopeTimebaseMode.XY,
            "ROLL" : OscilloscopeTimebaseMode.ROLL
        }
        if resp in modes:
            return modes[resp]
        else:
            raise CommunicationError_ProtocolViolation(f"Unknown timebase mode {resp} received from device")

    def _set_timebase_scale(self, scale):
        # The setable timebase scale depends on the model and the current
        # mode. Check if we are in range ...

        tbLimitsYT = {
            "MSO5354" : (1e-9, 1000),
            "MSO5204" : (2e-9, 1000),
            "MSO5102" : (5e-9, 1000),
            "MSO5104" : (5e-9, 1000),
            "MSO5072" : (5e-9, 1000),
            "MSO5074" : (5e-9, 1000)
        }

        if self._get_timebase_mode() == OscilloscopeTimebaseMode.ROLL:
            if (scale < 200e-3) or (scale > 1000.0):
                raise ValueError("Timebase values in roll mode have to be in range 200ms to 1ks")
        else:
            if self._id['product'] not in tbLimitsYT:
                raise ValueError("Failed to validate timebase region for unknown product {self._id['product']}")

            if (scale < tbLimitsYT[self._id['product']][0]) or (scale > tbLimitsYT[self._id['product']][1]):
                raise ValueError(f"Timebase scale {scale}s/div is out of range {tbLimitsYT[self._id['product']][0]}s/div to {tbLimitsYT[self._id['product']][1]}s/div for {self._id['product']}")

        # Set timebase
        self._scpi.scpiCommand(f":TIM:SCAL {scale}")

    def _get_timebase_scale(self):
        resp = self._scpi.scpiQuery(":TIM:SCAL?")
        try:
            resp = float(resp)
            return resp
        except:
            pass

        raise CommunicationError_ProtocolViolation(f"Unknown response for timebase scale: {resp}")

    def _set_channel_coupling(self, channel, mode):
        if (channel < 0) or (channel > 3):
            raise ValueError(f"Supplied channel number {channel} is out of bounds")

        modestr = {
            OscilloscopeCouplingMode.DC  : "DC",
            OscilloscopeCouplingMode.AC  : "AC",
            OscilloscopeCouplingMode.GND : "GND"
        }

        if mode not in modestr:
            raise ValueError(f"Unsupported coupling mode {mode}")

        self._scpi.scpiCommand(f":CHAN{channel+1}:COUP {modestr[mode]}")

    def _get_channel_coupling(self, channel):
        if (channel < 0) or (channel > 3):
            raise ValueError(f"Supplied channel number {channel} is out of bounds")

        resp = self._scpi.scpiQuery(f":CHAN{channel+1}:COUP?")

        modes = {
            "DC"   : OscilloscopeCouplingMode.DC,
            "AC"   : OscilloscopeCouplingMode.AC,
            "GND"  : OscilloscopeCouplingMode.GND
        }
        if resp in modes:
            return modes[resp]
        else:
            raise CommunicationError_ProtocolViolation(f"Unknown coupling mode {resp} received from device")

    def _set_channel_probe_ratio(self, channel, ratio):
        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Channel index {channel} is out of bounds")
        if ratio not in [ 0.0001, 0.0002, 0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000 ]:
            raise ValueError(f"Ratio {ratio} is not supported by this device")

        self._probe_ratios[channel] = ratio
        self._scpi.scpiCommand(f":CHAN{channel+1}:PROB {ratio}")

    def _get_channel_probe_ratio(self, channel):
        if (channel < 0) or (channel > 3):
            raise ValueError(f"Supplied channel number {channel} is out of bounds")
        resp = self._scpi.scpiQuery(f":CHAN{channel+1}:PROB?")

        try:
            resp = float(resp)
        except:
            return None

        if resp not in [ 0.0001, 0.0002, 0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000 ]:
            raise CommunicationError_ProtocolViolation(f"Received unsupported probe ratio {resp}")
        return resp

    def _set_channel_scale(self, channel, scale):
        if (channel < 0) or (channel > 3):
            raise ValueError(f"Supplied channel number {channel} is out of bounds")
        scale = float(scale)

        # Check it's a scale that's actually setable - one has to also look at channel probe ratio though ...
        setableScales = [ 500e-6, 1e-3, 2e-3, 5e-3, 1e-2, 2e-2, 5e-2, 1e-1, 2e-1, 5e-1, 1, 2, 5, 10 ]

        currentProbeRatio = self._get_channel_probe_ratio(channel)
        if currentProbeRatio is None:
            raise CommunicationError_ProtocolViolation("Failed to query current probe ratio")
        scale = scale / currentProbeRatio
        if scale not in setableScales:
            raise ValueError("Scale out of range [{500e-6 * currentProbeRatio};{10 * currentProbeRatio}] ({currentProbeRatio}x probe selected) in 1,2,5 steps")

        self._scpi.scpiCommand(f":CHAN{channel+1}:SCAL {scale}")

    def _get_channel_scale(self, channel):
        if (channel < 0) or (channel > 3):
            raise ValueError(f"Supplied channel number {channel} is out of bounds")
        resp = self._scpi.scpiQuery(f":CHAN{channel+1}:SCAL?")
        try:
            resp = float(resp)
        except:
            return None

        scalefactor = self._get_channel_probe_ratio(self, channel)
        if scalefactor is None:
            raise CommunicationError_ProtocolViolation("Failed to query current probe ratio")

        return resp * scalefactor

    def _waveform_get_xscale(self):
        xinc = self._scpi.scpiQuery(":WAV:XINC?")
        if xinc is None:
            raise CommunicationError_ProtocolViolation("Did not receive valid response to XINC")
        xorigin = self._scpi.scpiQuery(":WAV:XOR?")
        if xorigin is None:
            raise CommunicationError_ProtocolViolation("Did not receive valid response to XORIGIN")
        xref = self._scpi.scpiQuery(":WAV:XREF?")
        if xref is None:
            raise CommunicationError_ProtocolViolation("Did not receive valid response to XREF")

        try:
            xinc = float(xinc)
            xorigin = float(xorigin)
            xref = float(xref)
        except:
            raise CommunicationError_ProtocolViolation("Did not receive valid reply on XINC, XORIGIN or XREF")

        # This is:
        #	Interval between two neighboring points
        #	Start time of currently selected point after trigger
        #	Reference time (should be 0)
        return xinc, xorigin, xref

    def _waveform_get_yscale(self):
        xinc = self._scpi.scpiQuery(":WAV:YINC?")
        if xinc is None:
            raise CommunicationError_ProtocolViolation("Did not receive valid response to YINC")
        xorigin = self._scpi.scpiQuery(":WAV:YOR?")
        if xorigin is None:
            raise CommunicationError_ProtocolViolation("Did not receive valid response to YORIGIN")
        xref = self._scpi.scpiQuery(":WAV:YREF?")
        if xref is None:
            raise CommunicationError_ProtocolViolation("Did not receive valid response to YREF")

        try:
            xinc = float(xinc)
            xorigin = float(xorigin)
            xref = float(xref)
        except:
            raise CommunicationError_ProtocolViolation("Did not receive valid reply on YINC, YORIGIN or YREF")

        # This is:
        #	Interval between two neighboring points
        #	Start time of currently selected point after trigger
        #	Reference time (should be 0)
        return xinc, xorigin, xref

    def _query_waveform(self, channel, stats = None):
        if isinstance(channel, list) or isinstance(channel, tuple):
            resp = None

            for ch in channel:
                resp_next = self._query_waveform(ch)
                if resp is None:
                    resp = {
                        'x' : resp_next['x'],
                        f"y{ch}" : resp_next['y']
                    }
                else:
                    resp[f"y{ch}"] = resp_next['y']
            return resp

        if (channel < 0) or (channel >= self._nchannels):
            raise ValueError(f"Channel {channel} is out of range [0;{self._nchannels-1}]")
        self._scpi.scpiCommand(f":WAV:SOUR CHAN{channel+1}")
        self._scpi.scpiCommand(f":WAV:MODE NORM")
        self._scpi.scpiCommand(f":WAV:FORM ASCII")
        self._scpi.scpiCommand(f":WAV:POIN 1000")
        resppre = self._scpi.scpiQuery(":WAV:PRE?")
        respdata = self._scpi.scpiQuery(":WAV:DATA?")

        if (resppre is None) or (respdata is None):
            raise CommunicationError_ProtocolViolation("Failed to query trace from MSO5000")

        # Parse preamble ...
        pre = resppre.split(",")
        if len(pre) != 10:
            raise CommunicationError_ProtocolViolation("Unknown preamble format")

        if int(pre[0]) != 2:
            raise CommunicationError_ProtocolViolation(f"Requested ASCII but received format {pre[0]}")
        if int(pre[1]) != 0:
            raise CommunicationError_ProtocolViolation(f"Requested NORMAL data but received {pre[1]}")
        points = int(pre[2])
        avgcount = int(pre[3])
        xinc = float(pre[4])
        xorigin = float(pre[5])
        xref = float(pre[6])
        yinc = float(pre[7])
        yorigin = float(pre[8])
        yref = float(pre[9])

        # Parse data ...
        if respdata[0:2] != '#9':
            raise CommunicationError_ProtocolViolation("Trace data did not start with #9")
        wavebytes = int(respdata[2:11])
        wavedata = (respdata[11:wavebytes+11]).split(",")
        wavedata = [ float(i) for i in wavedata[:-1] ]

        # Build x axis ...
        if self._use_numpy:
            import numpy as np
            dpoints = len(wavedata)
            #xdata = np.arange(xorigin, xorigin + dpoints * xinc - 1e-9, xinc)
            xdata = np.linspace(xorigin, xorigin + dpoints * xinc, dpoints)
            ydata = np.asarray(wavedata)
        else:
            xdata = []
            curx = xorigin
            dpoints = len(wavedata)
            for i in range(dpoints):
                xdata.append(curx)
                curx = curx + xinc
            ydata = wavedata

        # Return trace X and Y axis ...
        #
        # The baseclass might add some statistics later on to the same dictionary

        res = {
            'x' : xdata,
            'y' : ydata
        }

        return res
